#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <algorithm>
#include "plugin.hpp"

#include "detector.hpp"
#include "types.hpp"
#include "tools.hpp"
//#include "lang.hpp"

void messageBoxDetectorIniCantOpen();
void messageBoxDetectorIniError();

static HANDLE ini;
static wchar_t iniBuf[256];
static DWORD  bufLenChars;
static DWORD  index;

static bool fgetch(wchar_t& ch)
{
  if(index == bufLenChars)
  {
    ULONG ReadSize;
    ReadFile(ini, iniBuf, sizeof(iniBuf), &ReadSize, NULL);
    bufLenChars = ReadSize / sizeof(wchar_t);
    if(bufLenChars == 0) return false;
    index = 0;
  }
  ch = iniBuf[index++];
  return true;
}

static bool fgets(wchar_t* to, int maxSize)
{
  static bool eof = false;
  if(eof) return false;

  int i = 0;
  for(; i < maxSize - 1; ++i)
  {
    wchar_t ch;
    if(!fgetch(ch))
    {
      eof = true;
      break;
    }
    if(ch == 0x0a) break;
    if(ch != 0x0d)
      to[i] = ch;
    else
      --i;
  }
  to[i] = 0;
  return true;
}

static wchar_t* skipWS(wchar_t* ptr)
{
  while(*ptr && (*ptr == L' ' || *ptr == L'\t')) ++ptr;
  return ptr;
}

static bool parseLine(wchar_t* ptr, wchar_t* key, wchar_t* value)
{
  while(*ptr && !(*ptr == L' ' || *ptr == L'\t' || *ptr == L'=')) *key++ = *ptr++;
  *key = 0;
  ptr = skipWS(ptr);
  if(!*ptr || *ptr != L'=') return false;
  ptr = skipWS(++ptr);
  if(!*ptr) return false;

  wchar_t* end = ptr + lstrlenW(ptr) - 1;
  while(end != ptr && (*end == L' ' || *end == L'\t')) --end; // skip end ws
  while(ptr != end+1) *value++ = *ptr++;
  *value = 0;
  return true;
}

static int atoi(wchar_t* ptr)
{
  int  result = -1;
  int  factor = 10;

  if(*ptr == L'#') { factor = 0x10; ++ptr;}
  if(*ptr == L'0' && *(ptr+1) == L'x') { factor = 0x10; ptr += 2; }
  
  while(1)
  {
    BYTE digit = 0xFF;
    if(*ptr >= L'0' && *ptr <= L'9') digit = *ptr - L'0';
    if(factor == 0x10)
    {
      if(*ptr >= L'A' && *ptr <= L'F') digit = *ptr - L'A' + 10;
      if(*ptr >= L'a' && *ptr <= L'f') digit = *ptr - L'a' + 10;
    }
    if(digit == 0xFF) break;
    if(result == -1) result = 0;
    
    result = factor*result + digit;
    ++ptr;
  }
  return result;
}

static bool getByte(BYTE& b, wchar_t* ptr)
{
  if(*ptr == L'\'' && *(ptr+2) == L'\'')
  {
    b = *(ptr+1);
    return true;
  }
  int num = atoi(ptr);
  if(num < 0 || num > 255) return false;
  b = num;
  return true;
}

static bool getWord(WORD& w, wchar_t* ptr)
{
  if(*ptr == L'\'' && *(ptr+3) == L'\'')
  {
    w = 256*(*(ptr+2)) + *(ptr+1);
    return true;
  }
  int num = atoi(ptr);
  if(num < 0 || num > 65535) return false;
  w = num;
  return true;
}

static bool checkFormat(const FormatInfo& form)
{
  if(form.type          ==  0  &&
     form.size          == -1 &&
     form.start         == -1 &&
     form.noSecs        == -1 &&
     form.signatureSize ==  0) return false;
  
  if(form.specialChar   == 0 &&
     form.newType       == 0 &&
     form.commentSize   == 0 &&
     form.description   == 0) return false;
  return true;
}

Detector::Detector(const wchar_t *iniFilePath, const wchar_t* path)
{
  wchar_t fileName[300];
  noFormats = 0;
  while(1)
  {
    if(iniFilePath[0] != 0)
    {
      lstrcpyW(fileName, iniFilePath);
      ini = CreateFileW(fileName,
                       GENERIC_READ,
                       FILE_SHARE_READ | FILE_SHARE_WRITE,
                       NULL,
                       OPEN_EXISTING,
                       FILE_FLAG_SEQUENTIAL_SCAN,
                       NULL);

      if(ini != INVALID_HANDLE_VALUE) break;
    }

    lstrcpyW(fileName, path);
    lstrcpyW(pointToName(fileName), L"types.ini");

    ini = CreateFileW(fileName,
                     GENERIC_READ,
                     FILE_SHARE_READ | FILE_SHARE_WRITE,
                     NULL,
                     OPEN_EXISTING,
                     FILE_FLAG_SEQUENTIAL_SCAN,
                     NULL);

    if(ini != INVALID_HANDLE_VALUE) break;

    lstrcpyW(fileName, path);
    wchar_t *ptr = pointToName(fileName) - 1;
    *ptr = 0;
    lstrcpyW(pointToName(fileName), L"types.ini");

    ini = CreateFileW(fileName,
                     GENERIC_READ,
                     FILE_SHARE_READ | FILE_SHARE_WRITE,
                     NULL,
                     OPEN_EXISTING,
                     FILE_FLAG_SEQUENTIAL_SCAN,
                     NULL);

    if(ini != INVALID_HANDLE_VALUE) break;

    messageBoxDetectorIniCantOpen();

    return;
  }

  index = bufLenChars = std::size(iniBuf);

  wchar_t  line [256];
  wchar_t  key  [20];
  wchar_t  value[128];
  wchar_t* ptr;
  
  bool errors = false;

  FormatInfo form;
  ZeroMemory(&form, sizeof(FormatInfo));
  form.size   = -1;
  form.start  = -1;
  form.noSecs = -1;
  
  while(fgets(line, sizeof(line)))
  {
    if(noFormats == 255) break;
    ptr = skipWS(line);
    
    //     
    if(!*ptr || *ptr == L'#' || *ptr == L';') continue;

    if(*ptr == L'[')
    {
      if(checkFormat(form)) formats[noFormats++] = form;
      
      ZeroMemory(&form, sizeof(FormatInfo));
      form.size   = -1;
      form.start  = -1;
      form.noSecs = -1;
      continue;
    }

    if(!parseLine(ptr, key, value))
    {
      errors = true;
      continue;
    }

    if(!lstrcmpW(key, L"Type"))
    {
      BYTE type;
      if(getByte(type, value))
        form.type = type;
      else
        errors = true;
      continue;
    }

    if(!lstrcmpW(key, L"Size"))
    {
      WORD size;
      if(getWord(size, value))
        form.size = size;
      else
        errors = true;
      continue;
    }

    if(!lstrcmpW(key, L"Start"))
    {
      WORD start;
      if(getWord(start, value))
        form.start = start;
      else
        errors = true;
      continue;
    }

    if(!lstrcmpW(key, L"NoSecs"))
    {
      BYTE noSecs;
      if(getByte(noSecs, value))
        form.noSecs = noSecs;
      else
        errors = true;
      continue;
    }

    if(!lstrcmpW(key, L"NewType"))
    {
      BYTE newType;
      if(getByte(newType, value))
        form.newType = newType;
      else
        errors = true;
      continue;
    }

    if(!lstrcmpW(key, L"SpecialChar"))
    {
      BYTE specialChar;
      if(getByte(specialChar, value))
        form.specialChar = specialChar;
      else
        errors = true;
      continue;
    }

    if(!lstrcmpW(key, L"Description"))
    {
      if(form.description) delete[] form.description;
      form.description = new wchar_t[lstrlenW(value)+1];
      wcscpy(form.description, value);
    }

    if(!lstrcmpW(key, L"ShowHeader"))
    {
      if(!lstrcmpiW(value, L"no")) form.skipHeader = true;
    }

    if(!lstrcmpW(key, L"Comment"))
    {
      WORD w;
      ptr = value;
      if(getWord(w, ptr))
        form.commentOffset = w;
      else
      {
        errors = true;
        continue;
      }
      while(*ptr && !(*ptr == L' ' || *ptr == L'\t')) ++ptr;
      ptr = skipWS(ptr);

      BYTE b;
      if(getByte(b, ptr))
        form.commentSize = b;
      else
        errors = true;
      continue;
    }

    if(!lstrcmpW(key, L"Signature"))
    {
      wchar_t* ptr = value;

      WORD w;
      if(getWord(w, value))
        form.signatureOffset = w;
      else
      {
        errors = true;
        continue;
      }
      int noQuestionMarks = 0;
      int i               = 0;
      
      for(; i < 32; ++i)
      {
        while(*ptr && !(*ptr == ' ' || *ptr == '\t')) ++ptr;
        ptr = skipWS(ptr);
        if(!*ptr) break; //  

        if(*ptr != '?')
        {
          BYTE b;
          if(!getByte(b, ptr))
          {
            errors = true;
            break;
          }
          form.signature[i] = b;
          form.signatureMask |= 1<<(32-i);
        }
        else
          ++noQuestionMarks;
      }
      if(noQuestionMarks < i) form.signatureSize = i;
    }
  }
  CloseHandle(ini);
  if(checkFormat(form) && noFormats != 255) formats[noFormats++] = form;
  if(errors)
  {
      messageBoxDetectorIniError();
  }
}

Detector::~Detector()
{
  for(int i = 0; i < noFormats; ++i)
    if(formats[i].description) delete[] formats[i].description;
}

BYTE Detector::detect(const SCLHdr& hdr, const BYTE* secs, int size, wchar_t* comment)
{
  if(noFormats == 0) return 0xFF;
  *comment = 0;
  
  for(BYTE i = 0; i < noFormats; ++i)
  {
    bool detected = false;
    BYTE mask = 0;
    if(formats[i].type   ==  0 || hdr.type   == formats[i].type)   mask |= 0x01;
    if(formats[i].size   == -1 || hdr.size   == formats[i].size)   mask |= 0x02;
    if(formats[i].start  == -1 || hdr.start  == formats[i].start)  mask |= 0x04;
    if(formats[i].noSecs == -1 || hdr.noSecs == formats[i].noSecs) mask |= 0x08;
    
    if(formats[i].signatureSize == 0)
      mask |= 0x10;
    else
    {
      if(formats[i].signatureOffset + formats[i].signatureSize < size)
      {
        int n = 0;
        for(; n < formats[i].signatureSize; ++n)
          if((formats[i].signatureMask & 1<<(32-n)) &&
             secs[formats[i].signatureOffset+n] != formats[i].signature[n]) break;
        if(n == formats[i].signatureSize) mask |= 0x10;
      }
    }
    if(mask == 0x1F) detected = true;
    if(detected)
    {
      if(formats[i].commentOffset + formats[i].commentSize < size)
      {
        if(formats[i].commentSize)
        {
          int j = 0;
          int commentSize = formats[i].commentSize;
          if(formats[i].commentSize > 99) commentSize = 99;
          for(; j < commentSize; ++j)
          {
            BYTE b = secs[formats[i].commentOffset+j];
            if(b > 31 && b < 128)
              comment[j] = b;
            else
              comment[j] = ' ';
          }
          comment[j] = 0;
          trim(comment);
        }
      }
      wchar_t* ptr = comment;
      if(*ptr)
      {
        int len = lstrlenW(ptr);
        wchar_t lastChar = L' ';
        
        for(int k = 0; k < len; ++k)
        {
          if(comment[k] != ' ')
          {
            *ptr++ = lastChar = comment[k];
          }
          else
          {
            if(lastChar != ' ') *ptr++ = lastChar = ' ';
          }
        }
        *ptr = 0;
      }
      return i;
    }
  }
  return 0xFF;
}
    
void Detector::specialChar(BYTE n, wchar_t *pos)
{
  if(n == 0xFF) return;
  if(formats[n].specialChar) *pos = formats[n].specialChar;
}

void Detector::getType(BYTE n, wchar_t *pos)
{
  if(n == 0xFF) return;
  if(formats[n].newType) *pos = formats[n].newType;
}

bool Detector::getSkipHeader(BYTE n)
{
  if(n == 0xFF) return 0;
  return formats[n].skipHeader;
}

const wchar_t* Detector::description(BYTE n)
{
  if(n == 0xFF) return 0;
  return formats[n].description;
}
