#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <string>

#include "plugin.hpp"
#include "types.hpp"
#include "trdos.hpp"
#include "tools.hpp"
#include "tools_xcreate.hpp"
#include "creator.hpp"
#include "lang.hpp"
#include "make.hpp"

extern PluginStartupInfo startupInfo;

static HANDLE image;

bool create(const wchar_t *fileName, const wchar_t *title, int writeProtect, int isDS, int format, int interleave, const wchar_t *bootName, const wchar_t *comment)
{
  BYTE   DSsignature[] = {'D', 'i', 'r', 'S', 'y', 's', '1', '0', '0'};
  int    noFiles   = 0;
  int    totalSecs = 0;
  HANDLE boot      = INVALID_HANDLE_VALUE;
  int    i;

  char cmt[256];

  //   cp866
  WideCharToMultiByte(866, 0, comment, 0, cmt, 0, nullptr, nullptr);

  DWORD noBytesRead;
  
  Track0 track0{};

  HANDLE screen = startupInfo.SaveScreen(0,0,-1,-1);

  if(bootName)
  {
    boot = CreateFileW(bootName,
                      GENERIC_READ,
                      FILE_SHARE_READ | FILE_SHARE_WRITE,
                      NULL,
                      OPEN_EXISTING,
                      FILE_FLAG_SEQUENTIAL_SCAN,
                      NULL);
    if(boot == INVALID_HANDLE_VALUE)
    {
      const wchar_t *msgItems[4];
      msgItems[0] = GetMsg(MError);
      msgItems[1] = GetMsg(MCanNotOpen);
      msgItems[2] = bootName;
      msgItems[3] = GetMsg(MOk);
      messageBox(FMSG_WARNING, msgItems, sizeof(msgItems)/sizeof(msgItems[0]), 1);
    }
    else
    {
      FileType t = detectFileType(boot);
      if(t == FMT_PLAIN)
      {
        const wchar_t *msgItems[4];
        msgItems[0] = GetMsg(MError);
        msgItems[1] = bootName;
        msgItems[2] = GetMsg(MUnknown);
        msgItems[3] = GetMsg(MOk);
        messageBox(FMSG_WARNING, msgItems, sizeof(msgItems)/sizeof(msgItems[0]), 1);
      }
      else
      {
        if(t == FMT_HOBETA)
        {
          HoHdr hdr;
          ReadFile(boot, &hdr, sizeof(HoHdr), &noBytesRead, 0);
          CopyMemory(&track0.files[0], &hdr, sizeof(HoHdr)-4);
          totalSecs = hdr.noSecs;
          track0.files[0].noSecs = hdr.noSecs;
          noFiles = 1;
        }
        else
        {
          SetFilePointer(boot, 8, NULL, FILE_BEGIN);
          BYTE no_files;
          ReadFile(boot, &no_files, 1, &noBytesRead, 0);
          for(i = 0; i < no_files; ++i)
          {
            SCLFileHdr hdr;
            ReadFile(boot, &hdr, sizeof(SCLFileHdr), &noBytesRead, 0);
            CopyMemory(&track0.files[i], &hdr, sizeof(SCLFileHdr));
            totalSecs += hdr.noSecs;
          }
          noFiles = no_files;
          if((format != FMTI_SCL) && ((noFiles > 128) || totalSecs > 2544))
          {
            const wchar_t *msgItems[4];
            msgItems[0] = GetMsg(MError);
            msgItems[1] = bootName;
            msgItems[2] = GetMsg(MTooBig);
            msgItems[3] = GetMsg(MOk);
            messageBox(FMSG_WARNING, msgItems, sizeof(msgItems)/sizeof(msgItems[0]), 1);
            noFiles = 0;
          }
        }
      }
    }
  }

  const wchar_t* ext = nullptr;
  switch(format)
  {
    case FMTI_TRD:
              ext = L".trd";
              break;
    case FMTI_FDI:
              ext = L".fdi";
              break;
    case FMTI_SCL:
              ext = L".scl";
              break;
    case FMTI_FDD:
              ext = L".fdd";
              break;
    case FMTI_UDI:
              ext = L".udi";
              break;
    case FMTI_TD:
              ext = L".td0";
              break;
  }

  std::wstring FileNameCopy{ fileName };
  if(!pointToExt(fileName)) FileNameCopy.append(ext);

  image = CreateFileW(FileNameCopy.c_str(),
                     GENERIC_READ | GENERIC_WRITE,
                     0,
                     NULL,
                     CREATE_NEW,
                     FILE_FLAG_SEQUENTIAL_SCAN,
                     NULL);
  if(image == INVALID_HANDLE_VALUE)
  {
    const wchar_t *msgItems[4];
    msgItems[0] = GetMsg(MError);
    msgItems[1] = GetMsg(MCanNotCreate);
    msgItems[2] = FileNameCopy.c_str();
    msgItems[3] = GetMsg(MOk);
    messageBox(FMSG_WARNING, msgItems, sizeof(msgItems)/sizeof(msgItems[0]), 1);
    return false;
  }

  if((format != FMTI_FDI) && writeProtect)
    SetFileAttributesW(FileNameCopy.c_str(), FILE_ATTRIBUTE_READONLY);

  const wchar_t* msgItems[2];
  msgItems[0] = NULL;
  msgItems[1] = GetMsg(MCreating);
  messageBox(0, msgItems, sizeof(msgItems)/sizeof(msgItems[0]), 0);

  int trk = 1;
  int sec = 0;
  int noDelFiles = 0;
  for(i = 0; i < noFiles; ++i)
  {
    track0.files[i].trk = trk;
    track0.files[i].sec = sec;
    if(track0.files[i].name[0] == 0x01) ++noDelFiles;
    trk += track0.files[i].noSecs/noSecs;
    sec += track0.files[i].noSecs%noSecs;
    if(sec > noSecs-1) { ++trk; sec -= noSecs; }
  }
  track0.flag         = 0x10;
  track0.type         = 0x16;
  track0.firstFreeTrk = trk;
  track0.firstFreeSec = sec;
  track0.noFiles      = noFiles;
  track0.noDelFiles   = noDelFiles;
  track0.noFreeSecs   = 2544 - totalSecs;

  std::fill(std::begin(track0.title), std::end(track0.title), ' ');
  for(i = 0; i < 11; ++i)
  {
    if(title[i] == 0) break;
    if(title[i] > 0) track0.title[i] = title[i];
  }

  if(isDS)
  {
    track0.DScrc = 0xD019;
    for(int i = 0; i < 9; ++i)
      track0.DSsignature[i] = DSsignature[i];
  }

  BYTE sectors[2][16] ={{1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16},
                        {1, 9, 2, 10, 3, 11, 4, 12, 5, 13, 6, 14, 7, 15, 8, 16}};

  switch(format)
  {
    case FMTI_TRD:
              createTRD(track0, totalSecs, image, boot);
              break;
    case FMTI_FDI:
              createFDI(track0, totalSecs, sectors[interleave-1], writeProtect, image, boot, cmt);
              break;
    case FMTI_SCL:
              createSCL(track0, totalSecs, image, boot);
              break;
    case FMTI_FDD:
              createFDD(track0, totalSecs, sectors[interleave-1], image, boot);
              break;
    case FMTI_UDI:
              createUDI(track0, totalSecs, sectors[interleave-1], image, boot, cmt);
              break;
    case FMTI_TD:
              createTD(track0, totalSecs, sectors[interleave-1], image, boot, cmt);
              break;
  }

  if(boot != INVALID_HANDLE_VALUE) CloseHandle(boot);
  CloseHandle(image);
  startupInfo.RestoreScreen(screen);
  return true;
}
