#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <iterator>

#include "types.hpp"
#include "types_isd.hpp"
#include "XiSD.hpp"
#include "manager.hpp"
#include "FmtReader.hpp"
#include "tools.hpp"
#include "lang.hpp"

PluginStartupInfo startupInfo;
FmtReader*        fmtReader;
Options           op;

static HANDLE Settings;

struct __declspec(uuid("{39772CEF-A39D-4D38-A509-303B03D7979A}")) TMainGuid {};
extern const GUID MainGuid = __uuidof(TMainGuid);

void WINAPI _export SetStartupInfoW(const struct PluginStartupInfo *info)
{
  //    PluginStartupInfo
  startupInfo = *info;
  fmtReader = new FmtReader(startupInfo.ModuleName);

  FarSettingsCreate fsc{ sizeof(FarSettingsCreate), MainGuid, 0 };
  startupInfo.SettingsControl(INVALID_HANDLE_VALUE, SCTL_CREATE, PSL_LOCAL, &fsc);
  Settings = fsc.Handle;

  FarSettingsItem fsi{ sizeof(FarSettingsItem), 0 };

  fsi.Name = L"DefaultPanelMode";
  fsi.Number = 4;
  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  op.defaultPanelMode = fsi.Number;

  op.reread = false;

  fsi.Type = FST_STRING;
  fsi.Name = L"ColumnTypes";
  fsi.String = L"N,S,C0,C1,C2";
  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  wcsncpy(op.columnTypes, fsi.String, std::size(op.columnTypes));
  op.columnTypes[std::size(op.columnTypes) - 1] = 0;

  fsi.Name = L"ColumnWidths";
  fsi.String = L"0,8,4,5,5";
  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  wcsncpy(op.columnWidths, fsi.String, std::size(op.columnWidths));
  op.columnWidths[std::size(op.columnWidths) - 1] = 0;

  fsi.Name = L"ColumnTitles";
  fsi.String = GetMsg(MTitle);
  startupInfo.SettingsControl(Settings, SCTL_GET, 0, &fsi);
  wcsncpy(op.columnTitles, fsi.String, std::size(op.columnTitles));
  op.columnTitles[std::size(op.columnTitles) - 1] = 0;
}

void WINAPI _export ExitFARW()
{
    startupInfo.SettingsControl(Settings, SCTL_FREE, 0, nullptr);
    Settings = nullptr;

    delete fmtReader;
    fmtReader = nullptr;
}

HANDLE WINAPI _export AnalyseW(const struct AnalyseInfo *Info)
{
  if(Info->FileName == nullptr) return nullptr;
  
  FmtPlugin* format = fmtReader->isImage(Info->FileName, (const BYTE *)Info->Buffer, Info->BufferSize);
  if(format == nullptr) return nullptr;
  
  Manager *m = new Manager(Info->FileName, format);
  return (HANDLE)m;
}

void WINAPI _export CloseAnalyseW(const struct CloseAnalyseInfo *Info)
{
    delete (Manager *)Info->Handle;
}

HANDLE WINAPI _export OpenW(const struct OpenInfo *Info)
{
  if(Info->OpenFrom == OPEN_ANALYSE)
  {
      OpenAnalyseInfo *An = (OpenAnalyseInfo *)Info->Data;
      return An->Handle;
  }

  if(Info->OpenFrom != OPEN_COMMANDLINE) return nullptr;
  const OpenCommandLineInfo *CommandLineInfo = (const OpenCommandLineInfo *)Info->Data;
  const wchar_t *fileName = CommandLineInfo->CommandLine;
  
  HANDLE hostFile = CreateFileW(fileName,
                               GENERIC_READ,
                               FILE_SHARE_READ | FILE_SHARE_WRITE,
                               NULL,
                               OPEN_EXISTING,
                               FILE_FLAG_SEQUENTIAL_SCAN,
                               NULL);
  if(hostFile == INVALID_HANDLE_VALUE) return INVALID_HANDLE_VALUE;
  
  BYTE data[blockSize];
  if(GetFileSize(hostFile, NULL) < sizeof(data))
  {
    CloseHandle(hostFile);
    return INVALID_HANDLE_VALUE;
  }

  DWORD noBytesRead;
  ReadFile(hostFile, data, sizeof(data), &noBytesRead, NULL);
  CloseHandle(hostFile);
  
  FmtPlugin* format = fmtReader->isImage(fileName, data, sizeof(data));
  if(format == NULL) return INVALID_HANDLE_VALUE;
  
  Manager *m = new Manager(fileName, format);
  return (HANDLE)m;
}

void WINAPI _export ClosePanelW(const struct ClosePanelInfo *Info)
{
    delete (Manager*)Info->hPanel;
}

void WINAPI _export GetOpenPanelInfoW(struct OpenPanelInfo *Info)
{
    Manager* m = (Manager*)Info->hPanel;
    m->GetOpenPanelInfo(Info);
}

struct __declspec(uuid("{B265E137-9323-40EB-82EF-6F2318149026}")) TMenuGuid {};
static const GUID MenuGuid = __uuidof(TMenuGuid);

void WINAPI _export GetPluginInfoW(PluginInfo *info)
{
  info->StructSize = sizeof(*info);
  info->Flags      = PF_DISABLEPANELS;

  static const wchar_t *cmdPrefix = L"XiSD";
  info->CommandPrefix = cmdPrefix;
  static const wchar_t *pluginCfgStrings[1];
  pluginCfgStrings[0] = GetMsg(MConfig);

  info->PluginMenu.Guids = &MenuGuid;
  info->PluginMenu.Strings = pluginCfgStrings;
  info->PluginMenu.Count = std::size(pluginCfgStrings);

  info->PluginConfig.Guids = &MenuGuid;
  info->PluginConfig.Strings = pluginCfgStrings;
  info->PluginConfig.Count = std::size(pluginCfgStrings);
}

intptr_t  WINAPI _export GetFindDataW(struct GetFindDataInfo *Info)
{
    Manager* m = (Manager *)Info->hPanel;
    return m->getFindData(&Info->PanelItem, &Info->ItemsNumber, Info->OpMode);
}

void WINAPI _export FreeFindDataW(const struct FreeFindDataInfo *Info)
{
    Manager* m = (Manager*)Info->hPanel;
    m->freeFindData(Info->PanelItem, Info->ItemsNumber);
}

intptr_t  WINAPI _export SetDirectoryW(const struct SetDirectoryInfo *Info)
{
    Manager* m = (Manager*)Info->hPanel;
    return m->setDirectory(Info->Dir, Info->OpMode);
}

intptr_t WINAPI _export GetFilesW(struct GetFilesInfo *Info)
{
    Manager* m = (Manager*)Info->hPanel;
    return m->getFiles(Info->PanelItem, Info->ItemsNumber, Info->Move, Info->DestPath, Info->OpMode);
}

intptr_t WINAPI _export DeleteFilesW(const struct DeleteFilesInfo *Info)
{
    Manager* m = (Manager*)Info->hPanel;
    return (m->deleteFiles(Info->PanelItem, Info->ItemsNumber, Info->OpMode));
}

intptr_t WINAPI _export ConfigureW(const struct ConfigureInfo *Info)
{
  wchar_t defaultPanelModeString[2] = L" ";
  defaultPanelModeString[0] = L'0' + op.defaultPanelMode;

  FarDialogItem dlgItems[] =
  {
    { DI_DOUBLEBOX, 3, 1, 50, 10, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MConfig) },
    { DI_TEXT,      5, 2,  0,  0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MDefPanelMode) },
    { DI_EDIT,     47, 2, 48,  2,  { TRUE }, nullptr, nullptr,                                   0, defaultPanelModeString },
    { DI_TEXT,      3, 3,  0,  0, { FALSE }, nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
    { DI_TEXT,      5, 4,  0,  0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MMode4Settings) },
    { DI_TEXT,      5, 5,  0,  0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MColumnTypes) },
    { DI_EDIT,     23, 5, 48,  5, { FALSE }, nullptr, nullptr,                                   0, op.columnTypes },
    { DI_TEXT,      5, 6,  0,  0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MColumnWidths) },
    { DI_EDIT,     23, 6, 48,  6, { FALSE }, nullptr, nullptr,                                   0, op.columnWidths },
    { DI_TEXT,      5, 7,  0,  0, { FALSE }, nullptr, nullptr,                                   0, GetMsg(MColumnTitles) },
    { DI_EDIT,     23, 7, 48,  7, { FALSE }, nullptr, nullptr,                                   0, op.columnTitles },
    { DI_TEXT,      3, 8,  0,  0, { FALSE }, nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
    { DI_BUTTON,    0, 9,  0,  0, { FALSE }, nullptr, nullptr, DIF_CENTERGROUP | DIF_DEFAULTBUTTON, GetMsg(MSave) },
    { DI_BUTTON,    0, 9,  0,  0, { FALSE }, nullptr, nullptr,                     DIF_CENTERGROUP, GetMsg(MCancel) }
  };

  std::unique_ptr<void, decltype(SafeDlgFree) *> Dlg{ startupInfo.DialogInit(&MainGuid,
                                   &__uuidof(struct __declspec(uuid("{59488958-0E5B-4A99-9F85-BE54DB2CB358}")) TDlgGuid),
                                   -1, -1, 54, 12,
                                   nullptr,
                                   dlgItems,
                                   std::size(dlgItems), 0, 0, startupInfo.DefDlgProc, nullptr), SafeDlgFree };
  if(Dlg.get() == INVALID_HANDLE_VALUE)
  {
      return FALSE;
  }
  intptr_t askCode = startupInfo.DialogRun(Dlg.get());

  if(askCode != 12) return FALSE;

  FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(defaultPanelModeString) - 1, defaultPanelModeString };
  startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 2, &ItemData);

  int mode = str2int(defaultPanelModeString, 1);
  if(mode < 0 || mode > 9) mode = 4;
  op.defaultPanelMode = mode;

  {
      FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(op.columnTypes) - 1, op.columnTypes };
      startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 6, &ItemData);
  }
  {
      FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(op.columnWidths) - 1, op.columnWidths };
      startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 8, &ItemData);
  }
  {
      FarDialogItemData ItemData = { sizeof(FarDialogItemData), std::size(op.columnTitles) - 1, op.columnTitles };
      startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 10, &ItemData);
  }

  FarSettingsItem fsi{ sizeof(FarSettingsItem), 0 };

  fsi.Type = FST_QWORD;

  fsi.Name = L"DefaultPanelMode";
  fsi.Number = op.defaultPanelMode;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Type = FST_STRING;

  fsi.Name = L"ColumnTypes";
  fsi.String = op.columnTypes;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"ColumnWidths";
  fsi.String = op.columnWidths;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  fsi.Name = L"ColumnTitles";
  fsi.String = op.columnTitles;
  startupInfo.SettingsControl(Settings, SCTL_SET, 0, &fsi);

  return TRUE;
}

intptr_t  WINAPI _export MakeDirectoryW(struct MakeDirectoryInfo *Info)
{
    Manager* m = (Manager*)Info->hPanel;
    return m->makeDirectory(Info->Name, Info->OpMode);
}

intptr_t WINAPI _export PutFilesW(const struct PutFilesInfo *Info)
{
    Manager* m = (Manager*)Info->hPanel;
    return (m->putFiles(Info->PanelItem, Info->ItemsNumber, Info->Move, Info->OpMode));
}

intptr_t WINAPI _export ProcessPanelInputW(const struct ProcessPanelInputInfo *Info)
{
    Manager* m = (Manager*)Info->hPanel;
    return m->processKey(&Info->Rec);
}

void WINAPI _export GetGlobalInfoW(struct GlobalInfo *Info)
{
    Info->StructSize = sizeof(struct GlobalInfo);
    Info->MinFarVersion = MAKEFARVERSION(3, 0, 0, 2927, VS_RELEASE);   //   Far Manager,    API
    Info->Version = MAKEFARVERSION(1, 0, 0, 1, VS_RC);   //   1.0.0.1, -
    Info->Guid = MainGuid;
    Info->Title = L"xISD";
    Info->Description = L"xISD";
    Info->Author = L"Elf/2 & Alexandr Medvedev";
}
