#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <iterator>

#include "plugin.hpp"

#include "types.hpp"
#include "types_scl.hpp"
#include "manager.hpp"
#include "tools.hpp"
#include "tools_scl.hpp"
#include "lang.hpp"

extern PluginStartupInfo startupInfo;
extern Options           op;

void addEndSlash(std::wstring &path);

int Manager::checkExistingFiles(SCLHdr* hdrs,
                                SCLHdr hdr,
                                size_t no_files,
                                int& allActions)
{
  //      
  bool fileAlreadyExists = false;
  int fNum;
  for(fNum = 0; fNum < no_files; ++fNum)
    if(memcmp(hdrs[fNum].name, hdr.name, 9) == 0)
    {
      fileAlreadyExists = true;
      break;
    }
  if(fileAlreadyExists)
  {
    if(allActions == 2) return 0;
    if(allActions == 0)
    {
      const wchar_t *msgItems[13];
      msgItems[0] = GetMsg(MWarning);
      msgItems[1] = GetMsg(MAlreadyExists);
      msgItems[3] = L"\x001";
      msgItems[4] = GetMsg(MLine);
      msgItems[7] = L"\x001";
      msgItems[8] = GetMsg(MWrite);
      msgItems[9] = GetMsg(MAll);
      msgItems[10] = GetMsg(MSkip);
      msgItems[11] = GetMsg(MSkipAll);
      msgItems[12] = GetMsg(MCancel);
      
      wchar_t name[13] = L"            ";
      makeTrDosName(name, hdr, op.showExt, 12);
      msgItems[2] = name;
      
      wchar_t source[50];
      wchar_t dest[50];
      wsprintfW(source, GetMsg(MSource),
               (int)hdr.size,
               hdr.noSecs);
      wsprintfW(dest,   GetMsg(MDestination),
               (int)hdrs[fNum].size,
               hdrs[fNum].noSecs);
      
      msgItems[5] = source;
      msgItems[6] = dest;

      int askCode = messageBox(FMSG_WARNING /*| FMSG_DOWN*/,
                               msgItems,
                               std::size(msgItems), 5);
      switch(askCode)
      {
        case -1:
        case 4:
                  return -1;
        case 3:
                  allActions = 2; //skipAll
        case 2:
                  return 0;
        case 1:
                  allActions = 1; //overwriteAll
      }
    }
  }
  return 1;
}

int Manager::putFiles(PluginPanelItem *panelItem, size_t noItems, int move, int opMode)
{
  if(noItems == 0) return 1;
  if(isReadOnly)
  {
      const wchar_t *msgItems[4];
    msgItems [0] = GetMsg(MWarning);
    msgItems [1] = GetMsg(MCanNotCopy);
    msgItems [2] = GetMsg(MWriteProtected);
    msgItems [3] = GetMsg(MOk);
    messageBox(FMSG_WARNING, msgItems, std::size(msgItems), 1);
    return FALSE;
  }

  wchar_t tempDirName[300];
  if(GetTempPathW(299, tempDirName) == 0) lstrcpyW(tempDirName, L".");

  wchar_t outFileName [300];
  wchar_t bodyFileName[300];
  
  if(GetTempFileNameW(tempDirName, L"out",  0, outFileName)  == 0) return 0;
  if(GetTempFileNameW(tempDirName, L"body", 0, bodyFileName) == 0) return 0;
  
  if(!openHostFile()) return 0;
  
  HANDLE bodyFile = CreateFileW(bodyFileName,
                               GENERIC_WRITE | GENERIC_READ,
                               FILE_SHARE_READ | FILE_SHARE_WRITE,
                               NULL,
                               CREATE_ALWAYS,
                               FILE_FLAG_SEQUENTIAL_SCAN | FILE_FLAG_DELETE_ON_CLOSE,
                               NULL);
  
  if(bodyFile == INVALID_HANDLE_VALUE)
  {
    closeHostFile();
    errorMessage(GetMsg(MCanNotCreateTmp));
    return 0;
  }
  
  HANDLE outFile = CreateFileW(outFileName,
                              GENERIC_WRITE,
                              FILE_SHARE_READ | FILE_SHARE_WRITE,
                              NULL,
                              CREATE_ALWAYS,
                              FILE_FLAG_SEQUENTIAL_SCAN,
                              NULL);
  
  if(outFile == INVALID_HANDLE_VALUE)
  {
    CloseHandle(bodyFile);
    closeHostFile();
    errorMessage(GetMsg(MCanNotCreateTmp));
    return 0;
  }
  
  DWORD   noBytesWritten, noBytesRead;
  SCLHdr hdrs[255];

  DWORD noSecsInHostFile = 0;   //     host 
  for(int i = 0; i < noFiles; ++i)
  {
    hdrs[i] = files[i];
    noSecsInHostFile += files[i].noSecs;
  }
  
  int   exitCode = 1;
  BYTE  totalFiles   = noFiles; //   
  DWORD totalSectors = 0;       //      

  int   allActions = 0;         //     write_all  skip_all

  HANDLE file;
  HANDLE screen = startupInfo.SaveScreen(0,0,-1,-1);

  size_t Size = startupInfo.PanelControl(PANEL_ACTIVE, FCTL_GETPANELDIRECTORY, 0, 0);
  std::unique_ptr<FarPanelDirectory, decltype(::free) *> dirInfo((FarPanelDirectory *)::malloc(Size), ::free);

  dirInfo->StructSize = sizeof(FarPanelDirectory);
  startupInfo.PanelControl(PANEL_ACTIVE, FCTL_GETPANELDIRECTORY, Size, dirInfo.get());

  for(int iNum = 0; iNum < noItems; ++iNum)
  {
      std::wstring FullPath(dirInfo->Name);
      addEndSlash(FullPath);
      FullPath.append(panelItem[iNum].FileName);

      const wchar_t *msgItems[2];
    msgItems[0] = GetMsg(MCopyingFile);
    msgItems[1] = pointToName(FullPath.c_str());
    if((opMode & OPM_SILENT) == 0)
      messageBox(0, msgItems, std::size(msgItems), 0);

    bool isFullFileCopied = true;
    
    file = CreateFileW(FullPath.c_str(),
                      GENERIC_READ,
                      FILE_SHARE_READ | FILE_SHARE_WRITE,
                      NULL,
                      OPEN_EXISTING,
                      FILE_FLAG_SEQUENTIAL_SCAN,
                      NULL);
    
    if(file == INVALID_HANDLE_VALUE)
    {
      wchar_t msg[300];
      wsprintfW(msg, GetMsg(MCanNotOpen), FullPath.c_str());
      errorMessage(msg);
      exitCode = 0;
      continue;
    }
    
    FileType file_type = detectFileType(file);
    if(file_type == FMT_HOBETA)
    {
      if(isSCLFull(totalFiles, 1, file))
      {
        exitCode = 0;
        break;
      }
      HoHdr ho_hdr;
      ReadFile(file, &ho_hdr, sizeof(HoHdr), &noBytesRead, 0);

      SCLHdr hdr;
      CopyMemory(hdr.name, ho_hdr.name, 8);
      hdr.type     = ho_hdr.type;
      hdr.start    = ho_hdr.start;
      hdr.size     = ho_hdr.size;
      hdr.noSecs   = ho_hdr.noSecs;
   
      int r = checkExistingFiles(hdrs, hdr, totalFiles, allActions);
      if(r == -1)   //  Cancel
      {
        exitCode = -1;
        break;
      }
      if(r == 0)    //  Skip/Skip all
      {
        CloseHandle(file);
        exitCode = 0;
        continue;
      }
      hdrs[totalFiles++] = hdr;
      
      copySectors(bodyFile, file, hdr.noSecs);

      totalSectors += hdr.noSecs;
    }
    if(file_type == FMT_SCL)
    {
      SetFilePointer(file, 8, NULL, FILE_BEGIN);
      BYTE no_files;
      ReadFile(file, &no_files, 1, &noBytesRead, 0);
      
      if(isSCLFull(totalFiles, no_files, file))
      {
        exitCode = 0;
        break;
      }

      int   totalSecs  = 0;
      DWORD fileOffset = no_files*sizeof(SCLHdr) + 8 + 1;
      
      for(int fNum = 0; fNum < no_files; ++fNum)
      {
        SCLHdr hdr;
        SetFilePointer(file, fNum*sizeof(SCLHdr) + 8 + 1, NULL, FILE_BEGIN);
        ReadFile(file, &hdr, sizeof(SCLHdr), &noBytesRead, NULL);

        int r = checkExistingFiles(hdrs, hdr, totalFiles, allActions);
        if(r == -1)   //  Cancel
        {
          exitCode = -1;
          break;
        }
        
        if(r == 0)  //  Skip/Skip all
        {
          isFullFileCopied = false;
          fileOffset += hdr.noSecs*sectorSize;
          exitCode = 0;
          continue;
        }

        hdrs[totalFiles++] = hdr;

        SetFilePointer(file, fileOffset, NULL, FILE_BEGIN);
        copySectors(bodyFile, file, hdr.noSecs);
        fileOffset += hdr.noSecs*sectorSize;
        
        totalSecs += hdr.noSecs;
      }
      totalSectors += totalSecs;
      if(exitCode == -1) break;
    }
    if(file_type == FMT_PLAIN)
    {
      auto fileSize  = panelItem[iNum].FileSize;
      int   sizeInSec = fileSize/sectorSize  + (fileSize%sectorSize ? 1 : 0);
      int   no_files  = sizeInSec/255 + (sizeInSec%255 ? 1: 0);
      if(!no_files) no_files = 1;
      
      if(isSCLFull(totalFiles, no_files, file))
      {
        exitCode = 0;
        break;
      }
      
      //  
      char name[8]  = {' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '};
      if(no_files > 1) FillMemory(name, 8, '_');
      
      const wchar_t* ptr = pointToName(FullPath.c_str());
      int pos = 0;
      while(*ptr && *ptr != '.')
      {
        name[pos++] = (*ptr < 0x80 && *ptr != 0x01) ? *ptr : '_';
        ++ptr;
        if(pos == 8) break;
      }
      
      //  
      char ext[3]   = {'C', 0, 0};
      ptr = pointToExt(FullPath.data());
      pos = 0;
      if(ptr)
      {
        while(*ptr)
        {
          ext[pos++] = (*ptr < 0x80) ? *ptr : '_';
          ++ptr;
          if(pos == 3) break;
        }
      }
      DWORD totalSecs = 0;
      for(int fNum = 0; fNum < no_files; ++fNum)
      {
        SCLHdr hdr;
        
        if(no_files > 1)
        {
          char num[5];
          wsprintf(num, "_%03d", fNum);
          CopyMemory(name + 4, num, 4);
        }
        
        CopyMemory(hdr.name, name, 8);
        CopyMemory(&hdr.type, ext,  3);
        if(sizeInSec > 255)
        {
          hdr.noSecs = 255;
          hdr.size   = 255 * sectorSize;
          
          sizeInSec -= 255;
          fileSize  -= 255 * sectorSize;
        }
        else
        {
          //  
          hdr.noSecs = sizeInSec;
          hdr.size   = fileSize;
        }

        int r = checkExistingFiles(hdrs, hdr, totalFiles, allActions);
        if(r == -1)   //  Cancel
        {
          exitCode = -1;
          break;
        }
        
        if(r == 0)  //  Skip/Skip all
        {
          isFullFileCopied = false;
          exitCode = 0;
          continue;
        }
        
        hdrs[totalFiles++] = hdr;
        
        SetFilePointer(file, fNum*255*sectorSize, NULL, FILE_BEGIN);
        copySectors(bodyFile, file, hdr.noSecs);
        
        totalSecs += hdr.noSecs;
      }
      if(exitCode == -1) break;
      totalSectors += totalSecs;
    }
    CloseHandle(file);
    file = INVALID_HANDLE_VALUE;
    if(isFullFileCopied)
    {
      panelItem[iNum].Flags ^= PPIF_SELECTED;
      if(move) DeleteFileW(FullPath.c_str());
    }
  }
  //      break
  if(file != INVALID_HANDLE_VALUE) CloseHandle(file);

  const wchar_t *msgItems[2];
  msgItems[0] = GetMsg(MCopy);
  msgItems[1] = GetMsg(MUpdateFiles);
  if((opMode & OPM_SILENT) == 0)
    messageBox(0, msgItems, std::size(msgItems), 0);

  static const BYTE  signature[] = {'S', 'I', 'N', 'C', 'L', 'A', 'I', 'R' };
  DWORD checkSum = 0x255;
  
  WriteFile(outFile, signature, sizeof(signature), &noBytesWritten, NULL);
  WriteFile(outFile, &totalFiles, 1, &noBytesWritten, NULL);
  checkSum += totalFiles;
  for(int i = 0; i < totalFiles; ++i)
  {
    WriteFile(outFile, &hdrs[i], sizeof(SCLHdr), &noBytesWritten, NULL);
    checkSum += calculateCheckSum((BYTE*)&hdrs[i], sizeof(SCLHdr));
  }

  DWORD firstFileOffset = noFiles*sizeof(SCLHdr) + 8 + 1;
  SetFilePointer(hostFile, firstFileOffset, NULL, FILE_BEGIN);
  checkSum += copySectors(outFile, hostFile, noSecsInHostFile);
  
  SetFilePointer(bodyFile, 0, NULL, FILE_BEGIN);
  checkSum += copySectors(outFile, bodyFile, totalSectors);
  
  WriteFile(outFile, &checkSum, 4, &noBytesWritten, NULL);
  
  CloseHandle(bodyFile);
  CloseHandle(outFile);
  closeHostFile();
  
  DeleteFileW(hostFileName);
  MoveFileW(outFileName, hostFileName);

  startupInfo.RestoreScreen(screen);
  return exitCode;
}
