#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <algorithm>

#include "plugin.hpp"
#include "types.hpp"
#include "FmtReader.hpp"
#include "tools.hpp"

extern PluginStartupInfo startupInfo;

FmtReader::FmtReader(const wchar_t *path)
{
  noPlugins = 0;
  wchar_t pluginPath[300];
  wchar_t pluginMask[300];
  
  lstrcpyW(pluginPath, path);
  lstrcpyW(pointToName(pluginPath), L"Formats\\");
  
  lstrcpyW(pluginMask, pluginPath);
  lstrcatW(pluginMask, L"*.fmt");

  WIN32_FIND_DATAW data;
  HANDLE foundFile = FindFirstFileW(pluginMask, &data);
  
  int done = (foundFile == INVALID_HANDLE_VALUE);
  while(!done)
  {
    wchar_t pluginName[300];
    lstrcpyW(pluginName, pluginPath);
    lstrcatW(pluginName, data.cFileName);

    HMODULE hPlugin = LoadLibraryW(pluginName);
    if(hPlugin != NULL)
    {
      
      plugins[noPlugins].handle  = hPlugin;
      plugins[noPlugins].isImage = (IS_IMAGE)GetProcAddress(hPlugin,"isImage");
      plugins[noPlugins].init    = (INIT)    GetProcAddress(hPlugin,"init");
      plugins[noPlugins].cleanup = (CLEANUP) GetProcAddress(hPlugin,"cleanup");
      plugins[noPlugins].reload  = (RELOAD)  GetProcAddress(hPlugin,"reload");
      plugins[noPlugins].open    = (OPEN)    GetProcAddress(hPlugin,"open");
      plugins[noPlugins].close   = (CLOSE)   GetProcAddress(hPlugin,"close");
      plugins[noPlugins].read    = (READ)    GetProcAddress(hPlugin,"read");
      plugins[noPlugins].write   = (WRITE)   GetProcAddress(hPlugin,"write");

      plugins[noPlugins].getFormatName = (GET_FORMAT_NAME)GetProcAddress(hPlugin,"getFormatName");
      plugins[noPlugins].isProtected   = (IS_PROTECTED)   GetProcAddress(hPlugin,"isProtected");
      plugins[noPlugins].protect       = (PROTECT)        GetProcAddress(hPlugin,"protect");
      
      if(plugins[noPlugins].isImage != NULL &&
         plugins[noPlugins].init    != NULL &&
         plugins[noPlugins].cleanup != NULL &&
         plugins[noPlugins].reload  != NULL &&
         plugins[noPlugins].open    != NULL &&
         plugins[noPlugins].close   != NULL &&
         plugins[noPlugins].read    != NULL &&
         plugins[noPlugins].write   != NULL &&
         plugins[noPlugins].getFormatName != NULL &&
         plugins[noPlugins].isProtected   != NULL &&
         plugins[noPlugins].protect       != NULL) ++noPlugins;
      if(noPlugins == 16) break;
    }
    done = !FindNextFileW(foundFile, &data);
  }
  FindClose(foundFile);
}

FmtReader::~FmtReader()
{
    std::for_each_n(std::cbegin(plugins), noPlugins, [](const FmtPlugin &p) { FreeLibrary(p.handle); });
}

const FmtPlugin* FmtReader::isImage(const wchar_t *fileName, const BYTE *data, int size)
{
    auto r = std::find_if(std::cbegin(plugins), std::cbegin(plugins) + noPlugins,
        [fileName, data, size](const FmtPlugin &p)
        {
            return p.isImage(fileName, data, size);
        }
    );
  return (r == std::cbegin(plugins) + noPlugins) ? nullptr : &*r;
}
